#include "server.h"

ServerConnection::ServerConnection(ServerConfiguration configuration) : configuration(configuration) { }

void ServerConnection::setCallbacks(ServerConnectionCallbacks *callbacks) {
    controller = callbacks;
}

void ServerConnection::shouldConnect(bool connect) {
    if (connect == shouldBeConnected) {
        return;
    }

    if (controller == NULL) {
        Serial.println("[ERROR] No callbacks set for server");
        return;
    }
    shouldBeConnected = connect;
    nextReconnectAttemptMs = currentTime;
}

void ServerConnection::connect() {
    Serial.printf("[INFO] Connecting to %s:%d%s\n", configuration.url, configuration.port, configuration.path);
    connectionTimeout = currentTime + configuration.socketHeartbeatIntervalMs;
    webSocket.begin(configuration.url, configuration.port, configuration.path);
    webSocket.setAuthorization(configuration.key);

    std::function<void(WStype_t, uint8_t *, size_t)> f = [this](WStype_t type, uint8_t *payload, size_t length) {
        this->webSocketEventHandler(type, payload, length);
    };
    webSocket.onEvent(f);
    webSocket.setReconnectInterval(configuration.reconnectTime);
}

void ServerConnection::didChangeConnectionState(bool isConnected) {
    static bool wasConnected = false;
    if (isConnected) {
        Serial.println("[INFO] Socket connected, enabling heartbeat");
        isConnecting = false;
        webSocket.enableHeartbeat(configuration.socketHeartbeatIntervalMs, configuration.socketHeartbeatTimeoutMs, configuration.socketHeartbeatFailureReconnectCount);
    } else {
        isDisconnecting = false;
        if (!wasConnected && shouldBeConnected && nextReconnectAttemptMs < currentTime) {
            nextReconnectAttemptMs = currentTime + configuration.socketHeartbeatIntervalMs;
            Serial.println("[INFO] Socket disconnected, setting reconnect time");
        } else if (wasConnected) {
            Serial.println("[INFO] Socket disconnected");
        }
    }
    wasConnected = isConnected;
}

void ServerConnection::loop(uint32_t millis) {
    currentTime = millis;
    webSocket.loop();

    if (shouldBeConnected) {
        if (isDisconnecting) {
            return; // Wait for disconnect to finish, then it will be reconnected
        }
        if (isConnecting) {
            if (millis > connectionTimeout) {
                // Cancel connection attempt
                Serial.println("[INFO] Canceling socket connection attempt");
                isDisconnecting = true;
                isConnecting = false;
                webSocket.disconnect();
            }
            return; // Wait for connect to finish
        }
        if (webSocket.isConnected()) {
            return; // Already connected
        }
        if (controller == NULL) {
            return;
        }
        if (millis < nextReconnectAttemptMs) {
            return; // Wait for next reconnect
        }
        isConnecting = true;
        connect();
    } else {
        if (isDisconnecting) {
            return; // Wait for disconnect
        }
        if (isConnecting) { 
            return; // Wait until connection is established, then it will be disconnected
        }
        if (!webSocket.isConnected()) {
            return;
        }
        isDisconnecting = true;
        Serial.println("[INFO] Disconnecting socket");
        webSocket.disconnect();
    }
}

void ServerConnection::webSocketEventHandler(WStype_t type, uint8_t * payload, size_t length) {
switch(type) {
    case WStype_DISCONNECTED:
        didChangeConnectionState(false);
        break;
    case WStype_CONNECTED:
        didChangeConnectionState(true);
        break;
    case WStype_TEXT:
        controller->sendServerError(MessageResult::TextReceivedOverSocket);
        break;
    case WStype_BIN:
        controller->handleServerMessage(payload, length);
        break;
    case WStype_PONG:
        break;
    case WStype_PING:
        break;
    case WStype_ERROR:
    case WStype_FRAGMENT_TEXT_START:
    case WStype_FRAGMENT_BIN_START:
    case WStype_FRAGMENT:
    case WStype_FRAGMENT_FIN:
    Serial.printf("[WARN] Unexpected socket event %d\n", type);
        controller->sendServerError(MessageResult::UnexpectedSocketEvent);
        break;
    }
}

void ServerConnection::sendResponse(uint8_t* buffer, uint16_t length) {
    if (socketIsConnected()) {
        webSocket.sendBIN(buffer, length);
    } else {
        Serial.println("Failed to send response, socket not connected.");
    }
}